// VinciaMGInterface.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2013 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Functions to interface to matrix elements generated with MadGraph
#ifndef VINCIA_MGInterface_H
#define VINCIA_MGInterface_H

// Include Pythia and headers
#include "Basics.h"
#include "Event.h"
#include "Info.h"
#include "ParticleData.h"
#include "PythiaStdlib.h"
#include "Settings.h"
#include "StandardModel.h"
//Used for ostream_iterator, which is used to convert a vector of integers
//into a vector of doubles
#include <iostream>
#include <iterator>

// Include Vincia Standard Library
#include "VinciaCommon.h"

// Stdlib header for complex numbers.
#include <complex>

using namespace std;
using namespace Pythia8;

// Function to initialize MadGraph Interface. 
// Returns false if MadGraph dummy library linked, true otherwise.

namespace Vincia {

  class MGInterface {

  public:

    // Constructor.
    MGInterface() { 
      isInitPtr=false; 
      isInit=false;
      isMassiveS=false;
      isMassiveC=false;
      isMassiveB=false;
    };
    
    // Destructor
    virtual ~MGInterface() {};

    // Set pointers to required PYTHIA 8 objects
    void initPtr(Couplings* couplingsPtrIn, ParticleData* particleDataPtrIn, 
		 Settings* settingsPtrIn, VinciaCommon* vinComPtrIn) {
      couplingsPtr    = couplingsPtrIn;
      particleDataPtr = particleDataPtrIn;
      settingsPtr     = settingsPtrIn;
      vinComPtr       = vinComPtrIn;
      isInitPtr       = true;
    }
    
    // Initialization routine. Must be called for anything to work.
    bool init();

    // Wrapper to set a parameter in a MadGraph common block
    void setParm(string which, double what);
    void setParm(string which, std::complex<double> what);

    // Wrapper to MadGraph matrix elements (without and with polarization)
    // (should return -1 if relevant ME not implemented)
    double ME2(int nIn, vector<int> id, vector<Vec4> p);
    // Madgraph uses integers for polarizations
    double ME2(int nIn, vector<int> id, vector<Vec4> p, vector<int> pol);
    // Wrapper to translate from double-precision polarizations
    double ME2(int nIn, vector<int> id, vector<Vec4> p, vector<double> pol){
      vector<int> polInt(pol.begin(), pol.end());    
      copy(pol.begin(), pol.end(), back_inserter(polInt));
      return ME2(nIn, id, p, polInt);    
    }
    
    // Set verbose level
    void setVerbose(int verboseIn) {verbose = verboseIn;}

    // Set color depth
    void setColorDepth(int colorDepthIn) {colorDepth = colorDepthIn;}
    // Get color depth
    int getColorDepth() {return colorDepth;}

    // Function to swap couplings (eg to use Zbb matrix element for Zcc)
    void swapZud();
    
  private:
    
    // Is initialized?
    bool isInitPtr, isInit;

    // Pointers to VINCIA and PYTHIA 8 objects
    Couplings*    couplingsPtr;
    ParticleData* particleDataPtr;
    Settings*     settingsPtr;
    VinciaCommon* vinComPtr;

    // Color Mode (0: leading color, 1: full color)
    int colorDepth;

    // Massive
    bool isMassiveS, isMassiveC, isMassiveB, isMassiveTau;
    
    // Verbose level
    int verbose;

  };

}
#endif

