// VinciaFunctions.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2013 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for the VinciaFunction class

#include "VinciaFunctions.h"

using namespace Pythia8;

namespace Vincia {

  //***********************************************************************
  // DGLAP splitting kernels with helicity-dependence
  //   (from Larkoski, Peskin: http://arxiv.org/abs/arXiv:0908.2450)
  // Note: Pg2gg is written the standard factor 2 normalization convention
  //       Gives factor 2 difference with respect PYTHIA FSR.
  //       The difference is related to whether P is interpreted as the
  //       probability for gluon A to branch to BC (without factor 2),
  //       or as the probability for gluon B to have originated from A
  //       (with factor 2). 
  //   Both quarks and gluons have definite helicities
  //      9: unpolarized
  //     +1: right-handed
  //     -1: left-handed
  //   z is momentum fraction taken by B in A -> BC
  //***********************************************************************

  // Pg2gg : written with the standard factor 2 normalization convention
  double Functions::Pg2gg(double z, int hA, int hB, int hC) {

    // If A unpolarized, treat as unpolarized
    if (hA == 9) return 2*pow2(1.-z*(1.-z))/z/(1.-z);

    // Expressions coded for hA = 1. Flip helicities if hA = -1.
    if (hA == -1) {
      hB *= -1;
      hC *= -1;
    }

    if (hB == 1 && hC == 1) return 1./z/(1.-z);
    else if (hB == -1 && hC == 1) return pow3(1.-z)/z;
    else if (hB == 1 && hC == -1) return pow3(z)/(1.-z);

    return 0.;
  }

  // Pg2qq : B is quark, C is antiquark
  // Note: mass corrections not implemented
  double Functions::Pg2qq(double z, int hA, int hB, int hC, double) {

    // If A unpolarized, treat as unpolarized
    if (hA == 9) return pow2(z) + pow2(1.-z);

    // Preserve quark helicity
    if (hB != -hC || abs(hB) != 1) return 0.;

    // Expressions coded for hA = 1. Flip helicities if hA = -1.
    if (hA == -1) {
      hB *= -1;
      hC *= -1;
    }

    if (hB == -1 && hC == 1) return pow2(1.-z);
    else if (hB == 1 && hC == -1) return pow2(z);

    return 0.;
  }

  // Pq2qg : B is quark, C is gluon
  // Note: mass corrections not implemented
  double Functions::Pq2qg(double z, int hA, int hB, int hC, double) {

    // If A unpolarized, treat as unpolarized
    if (hA == 9) return (1.+pow2(z))/(1.-z);

    // Preserve quark helicity
    if (hA != hB || abs(hB) != 1) return 0.;

    // Expressions coded for hA = +1. Flip helicities if hA = -1.
    if (hA == -1) {
      hB *= -1;
      hC *= -1;
    }

    if (hB == 1 && hC == -1) return pow2(z)/(1.-z);
    else if (hB == 1 && hC == 1) return 1./(1.-z);
    
    return 0.;
  }

  //***********************************************************************
  // DGLAP splitting kernels with polarization-dependence 
  //   (from Ellis, Stirling, Webber)
  //   Gluons are plane polarized, quarks have helicities
  //      9: unpolarized
  //     +1: in plane / right-handed
  //     -1: out-of-plane / left-handed
  //   z is momentum fraction taken by B in A -> BC
  //***********************************************************************

  // Pg2ggLin : written without factor 2 normalization convention
  double Functions::Pg2ggLin(double z, int polA, int polB, int polC) {

    // If A unpolarized, treat as unpolarized
    if (polA == 9) return (1.-z+pow2(z))/z/(1-z);

    // A in-plane polarized
    else if (polA == 1) {
      if (polB == 1 && polC == 1) return (1.-z)/z + z/(1.-z) + z*(1.-z);
      else if (polB == -1 && polC == -1) return z*(1.-z);
    }

    // A out-of-plane polarized
    else if (polA == -1) {
      if (polB == 1 && polC == -1) return (1.-z)/z;
      else if (polB == -1 && polC == 1) return z/(1.-z);
    }

    return 0.;
  }

  // Pg2qqLin : B is quark, C is antiquark
  // Note: mass corrections not implemented
  double Functions::Pg2qqLin(double z, int polA, int hB, int hC, double mu) {

    // If A unpolarized, treat as unpolarized
    if (polA == 9) return Pg2qq(z,9,9,9,mu);

    // Preserve quark helicity
    if (hB != -hC || abs(hB) != 1) return 0.;

    // A in-plane polarized
    else if (polA == 1) return pow2(1.-2*z);

    // A out-of-plane polarized 
    else if (polA == -1) return 1.;

    return 0.;
  }

  // Pq2qgLin : B is quark, C is gluon
  // Note: mass corrections not implemented
  double Functions::Pq2qgLin(double z, int hA, int hB, int polC, double mu) {

    // If A unpolarized, treat as unpolarized
    if (hA == 9) return Pq2qg(z,9,9,9,mu);

    // Preserve quark helicity
    if (hA != hB || abs(hB) != 1) return 0.;

    // Gluon polarized in plane
    else if (polC == 1) return pow2(1.+z)/(1.-z);

    // Gluon polarized out of plane
    else if (polC == -1) return 1.-z;
    
    return 0.;
  }

  //***********************************************************************
  // a30 (q_a, g_r, qbar_b) = Z->3 at tree level / |M_2|^2: LC
  // From GGG, JHEP 0509:056,2005 paper
  double Functions::A30(vector<Vec4>& pn, int a, int r, int b) {
    double sar = m2(pn[a],pn[r]);
    double srb = m2(pn[r],pn[b]);
    double sab = m2(pn[a],pn[b]);
    double s   = m2(pn[a],pn[r],pn[b]);
    // A30
    double A=(sar/srb+srb/sar+2*sab*s/sar/srb)/s;
    return A;
  }

  //***********************************************************************
  // a40 (q_a,g_b,g_c,qbar_d) = Z->4 at tree level / |M_2|^2: LC 
  // From GGG, JHEP 0509:056,2005 paper

  double Functions::A40(vector<Vec4>& pn, int a, int b, int c, int d) {
    double A = ( A40aux(pn,a,b,c,d) + A40aux(pn,d,c,b,a) );
    return A;
  }
  // Leading-color a40
  double Functions::A40aux(vector<Vec4>& pn, int a,int b,int c,int d) {
    // Compute invariants
    // Translate argument # to GGG notation
    double s12=m2(pn[a],pn[d]);
    double s13=m2(pn[a],pn[b]);
    double s14=m2(pn[a],pn[c]);
    double s23=m2(pn[d],pn[b]);
    double s24=m2(pn[d],pn[c]);
    double s34=m2(pn[b],pn[c]);
    double s134=m2(pn[a],pn[b],pn[c]);
    double s234=m2(pn[d],pn[b],pn[c]);
    double s1234=m2(pn[a],pn[b],pn[c],pn[d]);

    double A = 
      1.0/(2*s13*s24*s34) * (2*s12*s14 + 2*s12*s23 + 2*pow(s12,2) 
				      + pow(s14,2) + pow(s23,2) )
      + 1.0/(2*s13*s24*s134*s234)* (3*s12*pow(s34,2) - 4*pow(s12,2)*s34
				    + 2*pow(s12,3) - pow(s34,3))
      + 1.0/(s13*s24*s134) * (3*s12*s23 - 3*s12*s34 + 4*pow(s12,2)
			      - s23*s34 + pow(s23,2) + pow(s34,2) )
      + 3.0/(2*s13*s24) * (2*s12 + s14 + s23)
      + 1.0/(s13*s34) * (4*s12 + 3*s23 + 2*s24)
      + 1.0/(s13*pow(s134,2)) * (s12*s34 + s23*s34 + s24*s34)
      + 1.0/(s13*s134*s234) * (3*s12*s24 + 6*s12*s34 - 4*pow(s12,2)
			       - 3*s24*s34 - pow(s24,2) - 3*pow(s34,2) )
      + 1.0/(s13*s134) * ( -6*s12 - 3*s23 - s24 + 2*s34 )
      + 1.0/(s24*s34*s134) * ( 2*s12*s14 + 2*s12*s23 + 2*pow(s12,2) 
			       + 2*s14*s23 + pow(s14,2) + pow(s23,2) )
      + 1.0/(s24*s134) * ( -4*s12 - s14 - s23 + s34 )
      + 1.0/pow(s34,2) * ( s12 + 2*s13 - 2*s14 - s34 )
      + 1.0/(pow(s34,2)*pow(s134,2))  * ( 2*s12*pow(s14,2)
					  + 2*pow(s14,2)*(s23 + s24) )
      - (2.0*s12*s14*s24)/(pow(s34,2)*s134*s234) 
      + 1.0/(pow(s34,2)*s134) * (-2*s12*s14 - 4*s14*s24 + 2*pow(s14,2))
      + 1.0/(s34*s134*s234) * (- 2*s12*s14 - 4*pow(s12,2)
			       + 2*s14*s24 - pow(s14,2) - pow(s24,2) )
      + 1.0/(s34*s134) * (- 8*s12 - 2*s23 - 2*s24 )
      + 1.0/pow(s134,2)* (s12 + s23 + s24)
      + 3.0/(2*s134*s234) * (2*s12 + s14 - s24 - s34)
      + 1.0/(2*s134);
    
    // Normalize and return
    A /= s1234;
    return A;
  }

  //***********************************************************************
  // B40(q_a,q'_b,qbar'_c,qbar_d)
  // From GGG, JHEP 0509:056,2005
  double Functions::B40(vector<Vec4>& pn, int a, int b, int c, int d) {

    // B40 
    double A = B40aux(pn, a,b,c,d) + B40aux(pn,d,b,c,a) 
      + B40aux(pn,a,c,b,d) + B40aux(pn,d,c,b,a);

    return A;
  }

  // b40 
  // From GGG, JHEP 0509:056,2005
  double Functions::B40aux(vector<Vec4>& pn, int a, int b, int c, int d) {

    // Compute invariants
    // Translate argument # to GGG notation
    double s12=m2(pn[a],pn[d]);
    double s13=m2(pn[a],pn[b]);
    double s14=m2(pn[a],pn[c]);
    double s23=m2(pn[d],pn[b]);
    double s24=m2(pn[d],pn[c]);
    double s34=m2(pn[b],pn[c]);
    double s134=m2(pn[a],pn[b],pn[c]);
    double s234=m2(pn[d],pn[b],pn[c]);
    double s1234=m2(pn[a],pn[b],pn[c],pn[d]);
    
    // b40
    double A = 
      (s12*s13*s14+s13*s14*s23-pow2(s13)*s24)/pow2(s34)/pow2(s134)
      + (-s12*s13*s24+s13*s14*s23-s13*pow2(s24))/pow2(s34)/s134/s234
      + (s12*s13+s13*s23)/s34/pow2(s134)
      + (2.0*s12*s13+pow2(s12))/2.0/s34/s134/s234
      + s12/2.0/s134/s234;

    // Normalize and return
    A /= s1234;
    return A;

  }

}
