// vincia02.cc is a part of the VINCIA plugin to PYTHIA 8
// This is a simple test program to run WW production at LEP II
// Uses VINCIA for showering and matching and PYTHIA 8 for all else
// Copyright (c) 2013 Peter Skands

// PYTHIA and VINCIA include files
// (automatically defines and uses the Pythia and Vincia namespaces)
#include "Vincia.h"

// ROOT: include files for linking to ROOT, uncomment as necessary
// #include "TROOT.h"
// #include "TDirectory.h"
// #include "TFile.h"
// #include "TH1.h"
// #include "TH2.h"

// Main Program
int main() {

  //************************************************************************
  // Define Pythia 8 generator with plug-in

  Pythia pythia; 
  VinciaPlugin vincia(&pythia,"vincia02.cmnd");

  // Everything from here on is normal Pythia8 
  // (with some slight extra functionality, such as post-run statistics from 
  // Vincia and checking whether all events have unit weights)

  //************************************************************************

  // Shorthands
  Event& event = pythia.event;
  Settings& settings = pythia.settings;

  // Extract settings to be used in the main program.  
  int    nEvent    = settings.mode("Main:numberOfEvents");
  bool   vinciaOn  = settings.flag("Vincia");
  int    verbose   = settings.mode("Vincia:verbose");
  bool   helicityOn = settings.flag("Vincia:helicityShower");
  int    iEventPri = -1;

  //************************************************************************

  // Initialize
  pythia.init();

  //************************************************************************

  // Define a few PYTHIA utilities
  Thrust Thr(1);
  Sphericity SphLin(1,2);

  //************************************************************************

  // Define PYTHIA histograms
  // Rootlink: to use ROOT instead, HIST -> TH1F ("tag",...)

  // Number of quarks, partons, and charged particles
  double nQsum=0.0;
  double nPsum=0.0;
  double nCsum=0.0;
  Hist HnQuarks("nQuarks", 50, -0.5, 49.5);
  Hist HnPartons("nPartons", 100, -0.5, 99.5);
  Hist HnCharged("nCharged", 200, -1.0, 199.0);
  // Thrust
  int sBins=100;
  int sBinsCoarse=25;
  Hist Hthr("1-T",sBins, 0.0, 0.5);
  Hist HthrTail("1-T", sBinsCoarse, 0.0, 0.5);
  Hist HthrNorm("(1-T)*(1-T)", sBins, 0.0, 0.5);
  double wthr=sBins/0.5;
  double wthrCoarse=sBinsCoarse/0.5;
  // Event Shapes
  double wShape=sBins/1.0;
  Hist Hmaj("Major",sBins,0.0,1.0);
  Hist Hmin("Minor",sBins,0.0,1.0);
  Hist Hobl("Oblateness",sBins,0.0,1.0);
  Hist HistC("C",sBins,0.0,1.0);
  Hist HistD("D",sBins,0.0,1.0);
  Hist HistCNorm("C*(C)",sBins,0.0,1.0);
  Hist HistDNorm("D*(D)",sBins,0.0,1.0);

  //************************************************************************

  // Make a histogram showing alphaS for this run.

  Hist alphaS("alpha_S vs log10(Q)",100,-1.0,2.0);
  double dlogQ = 3.0/100.0;
  for (int iQ=0;iQ<100;iQ++) {
    double logQ = -1.0 + (iQ+0.5)*dlogQ;
    double Q    = pow(10,logQ);
    double as   = vincia.alphaS.alphaS(pow2(Q));
    if (as < 10.0) {
      if (verbose >= 5) cout<<" as("<<Q<<") = "<<as<<endl;
      alphaS.fill(logQ,as);
    }
  }

  //************************************************************************

  // EVENT GENERATION LOOP
  // Generation, event-by-event printout, analysis, and storage

  // (Optionally) re-initialize random number engine before event loop
  const int initseed = settings.mode("Random:seed");
  vincia.rndmPtr->init(initseed);

  // Event weight (default = 1)
  double evwt = 1.0;
  double sumW = 0.0;

  // Begin event loop
  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {

    // Verbose 
    if (iEvent >= iEventPri && iEventPri >= 0) {
      verbose = 8;
      vincia.setVerbose(verbose);
    }

    if (!pythia.next()) continue;

    if (iEvent == 0 || iEvent == iEventPri) 
      if (helicityOn) event.list(helicityOn);
    
    // Check for negative weights
    if (vinciaOn) evwt = vincia.weight();
    sumW += evwt;
    
    // Count FS charged hadrons, partons, and quarks
    int nc=0;
    int np=0;
    int nq=0;
    for (int i=1;i<event.size();i++) {
      // Count up final-state charged hadrons
      if (event[i].isCharged() && event[i].status() > 80) nc++;
      // Find last parton-level partons
      int iDau1 = event[i].daughter1();
      if (iDau1 == 0 || abs(event[iDau1].status()) > 80) {
	// Count up partons and quarks
	if (event[i].isQuark() || event[i].isGluon() ) np++;
	if (event[i].isQuark()) nq++;	
      }
    }
    HnQuarks.fill(0.5*nq,evwt);
    HnPartons.fill(1.0*np,evwt);
    HnCharged.fill(1.0*nc,evwt);
    nQsum += 0.5*nq;
    nPsum += np;
    nCsum += nc;

    // Histogram thrust 
    Thr.analyze( event );
    Hthr.fill(1.0-Thr.thrust(),wthr*evwt);
    HthrTail.fill(1.0-Thr.thrust(),wthrCoarse*evwt);    
    HthrNorm.fill(1.0-Thr.thrust(),(1-Thr.thrust())*wthr*evwt);    
    Hmaj.fill(Thr.tMajor(),wShape*evwt);
    Hmin.fill(Thr.tMinor(),wShape*evwt);
    Hobl.fill(Thr.oblateness(),wShape*evwt);
    
    // Histogram Linear Sphericity values
    if (np >= 2.0) {
      SphLin.analyze( event );
      double evC=3*(SphLin.eigenValue(1)*SphLin.eigenValue(2)
		    + SphLin.eigenValue(2)*SphLin.eigenValue(3)
		    + SphLin.eigenValue(3)*SphLin.eigenValue(1));
      double evD=27*SphLin.eigenValue(1)*SphLin.eigenValue(2)
	*SphLin.eigenValue(3);
      HistC.fill(evC,wShape*evwt);
      HistD.fill(evD,wShape*evwt);
      HistCNorm.fill(evC,evC*wShape*evwt);
      HistDNorm.fill(evD,evD*wShape*evwt);
    }

  }

  //************************************************************************

  // POST-RUN FINALIZATION 
  // Normalization, Statistics, Output

  //Normalize histograms to effective number of positive-weight events.
  double normFac = 1.0/sumW;

  Hthr *= normFac;
  HthrTail *= normFac;
  HthrNorm *= normFac;
  HistC *= normFac;
  HistD *= normFac;
  HistCNorm *= normFac;
  HistDNorm *= normFac;
  Hmaj *= normFac;
  Hmin *= normFac;
  Hobl *= normFac;

  HnQuarks    *= normFac;
  HnPartons   *= normFac;
  HnCharged   *= normFac;

  // Print a few histograms
  cout<<HnPartons<<endl;
  cout<<" ^nPartons"<<endl;
  cout<<Hthr<<endl;
  cout<<" ^(1-T)"<<endl;
  cout<<HistC<<endl;
  cout<<" ^C"<<endl;
  cout<<HistD<<endl;
  cout<<" ^D"<<endl;

  cout<<endl;
  cout<<fixed;
  cout<< " <nGluonSplit> = "<<num2str(nQsum * normFac-2.0)<<endl;
  cout<< " <nPartons>    = "<<num2str(nPsum * normFac)<<endl;
  cout<< " <nCharged>    = "<<num2str(nCsum * normFac)<<endl;
  cout<<endl;

  // Print out Vincia internal histograms
  if (vinciaOn && verbose >= 3) vincia.printHistos();

  // Print out Vincia end-of-run information
  pythia.statistics();
  if (vinciaOn) vincia.printInfo();

  //Output tables of histograms to file?   
  /*
  HThr.table(mycout);
  HC.table(mycout);
  HD.table(mycout);
  */

  // Done.
  return 0;
}
