// VinciaRoot.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2013 P. Skands
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the Vincia ROOT utilities

#ifndef VINCIAROOT_H
#define VINCIAROOT_H

// Include the VinciaRoot headers 
// (ROOT and STDLIB headers included in VinciaSettings.h)

#include "VRdisplay.h"
#include "VRfile.h"
#include "VRgraph.h"
#include "VRsettings.h"

// Put all the rest in Vincia namespace

namespace Vincia {
  
  //***************************************************************
  // class VinciaRoot: Main VinciaRoot Class
  // Note: is completely independent of VINCIA and can be used 
  // with any executable using ROOT histograms (specifically TH1F)
  //***************************************************************

  class VinciaRoot {

  public:

    // Constructor
    VinciaRoot() {

      // Switch on storage of stat uncertainties 
      TH1::SetDefaultSumw2();

    }
    
    // Initialize ROOT app
    void initApp() {
      cout<<"(VinciaRoot): initializing display app"<<endl;
      gROOT->Reset();
      gROOT->SetStyle("Plain");        
      TApplication *vcApp = new TApplication("vcApp",0,0);
      vcApp->InitializeGraphics();
      settings.isInitApp = true;
    }      

    // Check if a display exists
    bool exists(string displayName) {
      if (displays.find(displayName) == displays.end()) return false;
      else return true;
    }

    // Initialize a display
    void createDisplay(string displayName, int xSizIn=300, int ySizIn=400, 
                     int xOffsetIn=-1,int yOffsetIn=-1) {

      // Check if Root display app has been initialized
      if ( !settings.isInitApp ) initApp(); 

      // Check if display with that name already exists
      if ( exists(displayName) ) {
        cout<<"(VinciaRoot): warning! a display named "<<displayName
            <<" already exists"<<endl;
        displays.erase(displayName);
      }

      // Cascade windows from top right corner if no offsets specified
      if (xOffsetIn == -1) xOffsetIn = 16*displays.size();
      if (yOffsetIn == -1) yOffsetIn = 16*displays.size();

      // Create and initialize display 
      displays[displayName] = new VRdisplay(displayName,xSizIn,ySizIn,
                                            xOffsetIn,yOffsetIn);
      displays[displayName]->init(&settings);
 
    }

    // Create a new Display from TH1D*
    void createDisplay(string  displayName, TH1D* HisPtrIn, 
                       int xSizIn=300, int ySizIn=400, 
                       int xOffsetIn=-1,int yOffsetIn=-1) {
      // Initialize display and associate 
      createDisplay(displayName, xSizIn, ySizIn, xOffsetIn, yOffsetIn);
      displays[displayName]->addTH1DPtr(HisPtrIn);
    }
    
    // Create a new Display from vector<TH1D*>
    void createDisplay(string displayName, vector<TH1D*> HisPtrIn, 
                       int xSizIn=300, int ySizIn=400, 
                       int xOffsetIn=-1,int yOffsetIn=-1) {
      // Initialize display and associate 
      createDisplay(displayName, xSizIn, ySizIn, xOffsetIn, yOffsetIn);
      displays[displayName]->addTH1DPtr(HisPtrIn);
    }

    // (Re)set global verbose level
    void setVerbose(int verboseIn) {
      settings.verbose = verboseIn;
      // Propagate change to all displays
      for (map<string,VRdisplay*>::iterator it=displays.begin(); 
           it != displays.end(); ++it) it->second->setVerbose(verboseIn);
    }
    
    // (Re)set global optstat
    void setOptStat(int optStatIn) {
      settings.optStat = optStatIn;
      // Propagate change to all displays
      for (map<string,VRdisplay*>::iterator it=displays.begin(); 
           it != displays.end(); ++it) 
        it->second->setOptStat(optStatIn);
    }

    // Set global sigma multiplier 
    void setSigmaMC(double sigmaStatIn) {
      settings.sigmaStat = sigmaStatIn;
      // Propagate change to all displays
      for (map<string,VRdisplay*>::iterator it=displays.begin(); 
           it != displays.end(); ++it) 
        it->second->setSigmaMC(sigmaStatIn);
    }

    // Set normalize on/off
    void setNormalize(string displayNameIn, bool normalizeIn = true, 
                      double normFacIn = 1.0) {
      if (exists(displayNameIn)) displays[displayNameIn]->setNormalize(normalizeIn,
                                                                       normFacIn);
      else cout<<"(VinciaRoot): setNormalize no display named "<<displayNameIn<<endl;
    }

    // Set logarithmic Y axis
    void setLogY(string displayNameIn, bool logY=true) {
      if (exists(displayNameIn)) displays[displayNameIn]->setLogY(logY);
      else cout<<"(VinciaRoot): setLogY no display named "<<displayNameIn<<endl;
    }
    void setLogY(bool logY=true) {
      settings.logY = logY;
      // Propagate change to all displays
      for (map<string,VRdisplay*>::iterator it=displays.begin(); 
           it != displays.end(); ++it) 
        it->second->setLogY(logY);
    }
    // Set linear Y axis
    void setLinY(string displayNameIn, bool linY=true) { 
      if (exists(displayNameIn)) displays[displayNameIn]->setLogY(!linY); 
      else cout<<"(VinciaRoot): setLinY no display named "<<displayNameIn<<endl;
    }
    void setLinY(bool linY=true) { setLogY(!linY); }

    // Set MC names
    void setMCnames(string displayNameIn, string MCnamesIn) {
      if (exists(displayNameIn)) displays[displayNameIn]->setMCnames(MCnamesIn);
      else cout<<"(VinciaRoot): setMCnames no display named "<<displayNameIn
               <<endl;
    }
    void setMCnames(string MCnamesIn) {
      settings.MCnames = MCnamesIn;
      // Propagate change to all displays
      for (map<string,VRdisplay*>::iterator it=displays.begin(); 
           it != displays.end(); ++it) 
        it->second->setMCnames(MCnamesIn);
    }

    // Set legend (with optional alignment, default is auto)
    void setLegend(string displayNameIn, vector<string> legendIn, 
                   int alignIn = 0) {
      if (exists(displayNameIn)) 
        displays[displayNameIn]->setLegend(legendIn, alignIn);
      else cout<<"(VinciaRoot): setLegend no display named "<<displayNameIn
               <<endl;
    }
    void setLegend(vector<string> legendIn, int alignIn = 0) {
      settings.legend = legendIn;
      // Propagate change to all displays 
      for (map<string,VRdisplay*>::iterator it=displays.begin(); 
           it != displays.end(); ++it) 
        it->second->setLegend(legendIn, alignIn);
   }

    // Wrapper for passing single legend as string
    void setLegend(string displayNameIn, string legendIn) {
      vector<string> legendVec;
      legendVec.push_back(legendIn);
      setLegend(displayNameIn,legendVec);
    }
    void setLegend(string legendIn) {
      vector<string> legendVec;
      legendVec.push_back(legendIn);
      setLegend(legendVec);
    }

    // Set whether to show ratio pane by default.
    // Optionally loop over all existing plots as well
    void setShowRatio(bool showRatioIn=true, bool doForAll = true) { 
      settings.showRatioPad = showRatioIn;
      // Propagate change to all displays 
      if (doForAll) {
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) 
          it->second->setShowRatio(showRatioIn);
      }
    }
    // Wrapper to do the same for a single display
    void setShowRatio(string displayNameIn, bool showRatioIn) {
      if (exists(displayNameIn)) 
        displays[displayNameIn]->setShowRatio(showRatioIn);
      else cout<<"(VinciaRoot): setShowRatio no display named "<<displayNameIn
               <<endl;
    }
    
    // Functions to give access to member functions of individual displays
    void addTH1DPtr(string displayName, TH1D* hisPtrIn) {
      if ( !exists(displayName) ) createDisplay(displayName, hisPtrIn);
      else displays[displayName]->addTH1DPtr(hisPtrIn);
    }
    void addTH1DPtr(string displayName, vector<TH1D*> hisPtrIn) {
      if ( !exists(displayName) ) createDisplay(displayName, hisPtrIn);
      else displays[displayName]->addTH1DPtr(hisPtrIn);
    }    
    void setDataFile(string displayName, string fileIn, bool hasSysErrors=true) {
      if (hasSysErrors) setDataFile(displayName,fileIn,10,20,30,40,51,52,61,62);
      else setDataFile(displayName,fileIn,10,20,30,40,51,52);
    }
    void setDataFile(string displayName, string fileIn, 
                     int i1, int i2, int i3=-1,
                     int i4=-1, int i5=-1, int i6=-1, int i7=-1, int i8=-1, 
                     int i9=-1, int i10=-1, int i11=-1, int i12=-1) {
      if ( !exists(displayName) && settings.verbose >= 1) {
        cout<<"(VinciaRoot): ERROR! display "<<displayName
            <<" does not exist. Attempt to add data file failed."<<endl;
        return;
      }
      displays[displayName]->setDataFile(fileIn, i1, i2, i3, i4, i5, i6, i7,
                                         i8, i9, i10, i11, i12);
    }    
    void addHisFile(string displayName, string fileIn,  
                    bool hasSysErrors=false) {
      if (hasSysErrors) addHisFile(displayName,fileIn,10,20,30,40,51,52,61,62);
      else addHisFile(displayName,fileIn,10,20,30,40,51,52);
    }
    void addHisFile(string displayName,string fileIn, int i1, int i2, int i3=-1,
                     int i4=-1, int i5=-1, int i6=-1, int i7=-1, int i8=-1, 
                     int i9=-1, int i10=-1, int i11=-1, int i12=-1) {
      if ( !exists(displayName) && settings.verbose >= 1) {
        cout<<"(VinciaRoot): ERROR! display "<<displayName
            <<" does not exist. Attempt to add histogram file failed."<<endl;
        return;
      }
      displays[displayName]->addHisFile(fileIn, i1, i2, i3, i4, i5, i6, i7,
                                         i8, i9, i10, i11, i12);
    }    
    
    // Add objects
    void addTLine(string displayName, TLine* linePtr) {
      if ( exists(displayName) ) displays[displayName]->addTLine(linePtr);
    }

    // Save Main histograms (with uncertainties)
    void saveHistograms(string baseName, string fileType) {
      for (map<string,VRdisplay*>::iterator it=displays.begin(); 
           it != displays.end(); ++it) it->second->saveHistogram(baseName, 
                                                                 fileType);
    }

    // Update display(s): default is update all
    void update(string displayName="all") {
      if (displayName == "all") {
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) it->second->update();        
      } else if ( exists(displayName) ) {
        displays[displayName]->update();        
      }
    }
    // Legacy functions for backwards compatibility
    void updateDisplay(string displayName="all") { update(displayName); }
    void updateDisplays() { update(); }

    // Save Display
    void saveDisplay(string displayNameIn, 
                     string baseName, string fileType="png") {
      if (exists(displayNameIn)) displays[displayNameIn]->saveDisplay(baseName,
                                                                    fileType);
      else cout<<"(VinciaRoot): saveDisplay no display named "<<displayNameIn
               <<endl;
    }
    
    void saveDisplays(string baseName, string fileType="png") {
      for (map<string,VRdisplay*>::iterator it=displays.begin(); 
           it != displays.end(); ++it) it->second->saveDisplay(baseName, 
                                                               fileType);
    }

    // Return pointer to an entire display (if desired)
    VRdisplay* getDisplay(string displayName) { 
      if (exists(displayName)) return displays[displayName];
      else {
        cout<<"(VinciaRoot): ERROR! attempt to access non-existing display "
            <<displayName<<" - retuning NULL"<<endl;
        return NULL;
      }
    }

    // Save Display
    void printChi2() {
      for (map<string,VRdisplay*>::iterator it=displays.begin(); 
           it != displays.end(); ++it) it->second->printChi2();
    }
    
  private:

    // vector of displays
    map<string,VRdisplay*> displays;

    // Global settings
    VRsettings settings;

  };

  //***************************************************************
  // class UncertaintyBand: 
  // Extremely simple auxiliary class for storing names and indices 
  // of uncertainty bands
  //***************************************************************

  class UncertaintyBand {
    
  public: 
    
    // Constructor
    UncertaintyBand(string nameIn=" ") {
      name = nameIn;
      indices.resize(0);
    };
    
    // Name
    string name;
    // Indices
    vector<int> indices;
    
  };
 
}

#endif
